package peterson;

import java.io.IOException;

/**
 * Deze klasse regelt het hele electie-proces.
 * @author Pieter van Ede - 3019764 <br /> Wesley Kok - 3019853
 *
 */
public class Electie 
{
	private int id, poort;
	private final int poortBegin;
	private Semafoor beschikbareKanalen;
	private Queue kanalen;
	private Kanaal volgende, vorige;
	
	/**
	 * Maakt een nieuw electie-object, met alle benodigde informatie zodat er zometeen kan worden
	 * begonnen met de electie.
	 * 
	 * @param poort  poort op de localhost waarop dit proces bereikbaar is
	 * @param poortOpvolger  poort op de localhost van de rechterbuur in de originele ringconfiguratie
	 * @param id  identitiet van dit proces
	 * @param beschikbareKanalen  semafoor die aangeeft of er kanalen in de queue zitten
	 * @see Semafoor
	 * @param kanalen  de queue waarin de serverthread kanalen kan stoppen, en waaruit
	 * de newClient methode kanalen kan ophalen
	 * @see Queue
	 */
	public Electie(int poort, int poortOpvolger, int id, Semafoor beschikbareKanalen, Queue kanalen)
	{
		this.id = id;
		this.poort = poort;
		this.poortBegin = poortOpvolger;
		this.beschikbareKanalen = beschikbareKanalen;
		this.kanalen = kanalen;
		nieuweVerbindingen(poortOpvolger);
	}
	
	/**
	 * Wacht tot de serverthread een kanaal met een (nieuwe)linkerbuur in de queue heeft gezet,
	 * en haalt dit kanaal dan uit de queue en levert hem op.
	 * 
	 * @return  een (nieuwe) verbinding met een linkerbuur
	 */
	private Kanaal newClient()
	{
		beschikbareKanalen.p();
		return kanalen.dequeue();
	}
	
	/**
	 * Deze methode maakt nieuwe kanalen met de linker en rechterbuur.
	 * 
	 * @param poortOpvolger  de nieuwe rechterbuur waarmee kanalen moeten worden gemaakt
	 */
	private void nieuweVerbindingen(int poortOpvolger)
	{
		try
		{
			//sluiten van kanalen
			if (volgende != null)
				volgende.sluitKanaal();
			if (vorige != null)
				vorige.sluitKanaal();
			
			//We maken nu een nieuwe verbinding met de rechterbuur aan.
			ClientConnectionMaker clientMaker = new ClientConnectionMaker(poortOpvolger);
			
			//Zet kanaal op met de rechterbuur.
			volgende = clientMaker.nieuweBuur();
			
			//Accepteer verbinding met (nieuwe) linkerbuur, en zet een kanaal naar de 
			//rechter buur op.
			vorige = newClient();			
		}
		catch( IOException e) 
		{ 
			Util.printDebug(id + ": Error in maken van kanalen met linker- en rechterbuur", false);
		}
	}
	
	/**
	 * Het daadwerkelijke Peterson-electie algoritme, met shortcuts. Deze kiest een leider, en zorgt
	 * ervoor dat via print-statements het verkiezingsproces te volgen is.
	 * 
	 * @throws IOException
	 * @throws ClassNotFoundException
	 */
	public void startElectie() throws IOException, ClassNotFoundException
	{
		boolean passief = false;
		boolean leider = false;
		
		// de ronden voor niet passieve rakkers
		while(!passief && !leider)
		{
			//je idnummer sturen
			volgende.stuur(new Pakketje(id));
			vorige.stuur(new Pakketje(id));
			
			//id's van de buren ontvangen
			int linksid = vorige.ontvang().openPakketje();
			int rechtsid = volgende.ontvang().openPakketje();
			if(id > linksid || id > rechtsid)
			{	
				System.out.println(id + ": Ik word passief, want mijn buren zijn " + linksid 
						+ ", " + rechtsid + ".");
				passief = true;
			}
			else
			{
				//Je gaat een ronde door, en je geeft je eigen poort naar links door
				vorige.stuur(new Pakketje(poort));
				
				//ontvang het poortnummer van rechts					
				int rechtspoort = volgende.ontvang().openPakketje();
				
				//als je poort en rechtspoort gelijk zijn, ben je de enige nog in de electie en 
				//word je dus leider.
				if (poort == rechtspoort)
				{
					System.out.println(id + ": Ik ben leider.");
					leider = true;
				}
				else
				{
					System.out.println(id + ": Ik blijf actief, want mijn buren zijn " 
							+ linksid + ", " + rechtsid + ".");
					
					//Nieuwe connecties maken met nieuwe buren
					nieuweVerbindingen(rechtspoort);
				}
			}
		}
		
		//Peterson electie voor jou is afgelopen.
		int idleider;
		if (passief)
		{
			//poort van rechts naar links doorsturen, om shortcuts tot stand te brengen
			int rechts = volgende.ontvang().openPakketje();
			vorige.stuur(new Pakketje(rechts));
			
			//sluiten van kanalen
			volgende.sluitKanaal();
			vorige.sluitKanaal();
			
			//wachten tot je de id van de leider krijgt (dit gebeurt doordat newClient() pas
			//doorgaat, zodra er een nieuwe client-aanvraag is)
			//nu maken we een kanaal naar de vorige buur, zodat we het leider-id kunnen 
			//ontvangen
			vorige = newClient();
					
			//lees het leider-id en print die
			idleider = vorige.ontvang().openPakketje();
			System.out.println(id + ": De leider is: " + idleider);
		}
		else
		{
			//sluit even je oude volgende buren
			volgende.sluitKanaal();
			vorige.sluitKanaal();
			idleider = this.id;
		}
		
		//stuur jouw id of de gekregen id door (afhankelijk van of je leider of 
		// niet-leider bent)
		ClientConnectionMaker clientMaker = new ClientConnectionMaker(this.poortBegin);
		volgende = clientMaker.eenrichtingBuur();
		volgende.stuur(new Pakketje(idleider));
		volgende.sluitKanaal();	
		
		//als je de leider bent, wacht je even en controleer dat iedereen weet dat je 
		// leider bent
		if(leider)
		{
			vorige = newClient();
			idleider = vorige.ontvang().openPakketje();
			vorige.sluitKanaal();
			if (idleider == this.id)
				System.out.println(id + ": En ik weet dat iedereen weet dat ik weet dat ik de leider ben.");
		}			
	}
}